////////////////////////////////////////////////////////////
//                                                        //
// VRLab Mental Vision Module 7: Sweeping Techniques      //
// Author: Achille Peternier                              //
// Copyright: EPFL-VRLab 2005 (C)                         //
//                                                        //
////////////////////////////////////////////////////////////
#include "../data/common.h"
#include <GL/glu.h>
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")



//////////////
// #DEFINES //
//////////////
#define MODULE_NAME "Sweeping techniques"
#define MAX_POINTS 32
#ifndef WIN32
   #define CALLBACK
   #define __stdcall
#endif



/////////////////
// Tesselator: //
void CALLBACK beginCallback(GLenum which)
{
   glBegin(which);
}

void CALLBACK endCallback(void)
{
   glEnd();
}
bool invalid = false;
void CALLBACK errorCallback(GLenum errorCode)
{
   const GLubyte * estring;
   estring = gluErrorString(errorCode);
   //exit(0);
   invalid = true;
}



///////////////////////////////
// ?: build a sweeping solid //
bool createSurface(MVVECTOR2 *point, MVVECTOR3 **gpoint, int numberOfPoints, int steps, float angle, float delta)
{
   // Check params:
   if (point == NULL || numberOfPoints < 2 || steps < 1)
         return false;
         
   // Free memory?
   if (*gpoint != NULL)
         delete [] *gpoint;
         
   // Ok, allocate memory:
   float r = 7.0f;
   MVVECTOR3 *_gpoint = new MVVECTOR3[(numberOfPoints-1)*4*(steps+1)];
   for (int c=0; c<numberOfPoints; c++)   
          _gpoint[c] = MVVECTOR3((point[c].x - delta)/r, -point[c].y/r+55.0f, (float) 0.0f);    
   float angleStep = angle / (float) (steps-1);
   
   // Construct geometry:
   for (int s=1; s<steps; s++)
   {
         for (int c=0; c<numberOfPoints; c++)
         {               
               double alpha = E3M_TORAD(angleStep);
               float dx = _gpoint[(s-1)*numberOfPoints+c].x;
               float dz = _gpoint[(s-1)*numberOfPoints+c].z;
               float x = dx*cos(alpha) - dz*sin(alpha);
               float z = dx*sin(alpha) + dz*cos(alpha);                                           
               _gpoint[s*numberOfPoints+c] = MVVECTOR3(x, _gpoint[(s-1)*numberOfPoints+c].y, z);
         }
   }
   
   // Done:
   *gpoint = _gpoint;
   return true;
}



//////////
// MAIN //
//////////

int main(int argc, char *argv[])
{
   /////////////////
   // Init MVisio //
   /////////////////   
   INIT_MVISIO(MODULE_NAME, argc, argv);  
   
   // Common interface:
   USE_FONT();  
   USE_BACKGROUND(MODULE_SIZEX, MODULE_SIZEY);
   USE_QUITBUTTON(MODULE_SIZEX, MODULE_SIZEY);
   USE_CAMERA(0.0f, 0.0f, 300.0f, 0.0f, 0.0f, 0.0f);
   USE_LIGHT(0.0f, 10.0f, 0.0f);
   
   // 2 viewport corners:
   int left_sx, left_sy, left_ex, left_ey;
   int right_sx, right_sy, right_ex, right_ey;
   left_sx = 0.0f; left_sy = MODULE_SIZEY/17.0f;
   left_ex = MODULE_SIZEX/2.0f; left_ey = MODULE_SIZEY-MODULE_SIZEY/8.1f;   
   right_sx = MODULE_SIZEX/2.0f; right_sy = MODULE_SIZEY/17.0f;
   right_ex = MODULE_SIZEX; right_ey = MODULE_SIZEY-MODULE_SIZEY/8.1f;
   
   // Point list:
   MVVECTOR2 point[MAX_POINTS];
   EXF_ZEROMEMORY(point, sizeof(point));
   int usedPoints = 0;
   int oldUsedPoints = 0;
   int oldSteps = 0;
   float oldScrollValue = 0.0f;
   
   // Geometry list:
   MVVECTOR3 *gpoint = NULL;
   
   
   ////////////////////////////
   // Point modifier interface:
   MVWINDOW *interfaceWin = new MVWINDOW();
   interfaceWin->setFont(font);
   interfaceWin->setSizes(253, 140);  
   interfaceWin->setCaption("Interface");
   interfaceWin->setTexture(baseTexture);
   interfaceWin->setPosition((int) (right_ex-interfaceWin->getSizeX()*1.5f), (int) (right_ey-interfaceWin->getSizeY()*1.5f));
   
   // Reset:
   MVBUTTON *reset = new MVBUTTON();   
   reset->setFont(font);
   reset->setCaption("^x0007reset");
   reset->setPosition(8, 30);   
   reset->setSizes(50, 20);
   reset->link(interfaceWin);
   
   // Apply:
   MVBUTTON *apply = new MVBUTTON();   
   apply->setFont(font);
   apply->setCaption("^x0008apply");
   apply->setPosition(64, 30);   
   apply->setSizes(50, 20);
   apply->link(interfaceWin);   

   // Rotate:   
   MVBUTTON *rotate = new MVBUTTON();   
   rotate->setFont(font);
   rotate->setCaption("^x0004rotate");
   rotate->setPosition(120, 30);   
   rotate->setSizes(50, 20);
   rotate->setEventType(MV_BUTTON_SWITCH);
   rotate->link(interfaceWin);
   
   // Wireframe:   
   MVBUTTON *wireframe = new MVBUTTON();   
   wireframe->setFont(font);
   wireframe->setCaption("^x0007wireframe");
   wireframe->setPosition(175, 30);   
   wireframe->setSizes(70, 20);
   wireframe->setEventType(MV_BUTTON_SWITCH);
   wireframe->link(interfaceWin);
   
   // Plane:
   MVBUTTON *plane = new MVBUTTON();      
   plane->setPosition(MODULE_SIZEX/4, left_sy);   
   plane->setSizes(3, left_ey-left_sy);  
   plane->setThickness(1);
   plane->setStaticStatus(false);
   plane->setLimitStatus(true);
   plane->setColor(1.0f, 1.0f, 1.0f, 0.5f);
   plane->setLimits(left_sx, left_sy, left_ex, left_ey);
   
   // Angles:
   MVHSCROLLBAR *scroll1 = new MVHSCROLLBAR();   
   scroll1->setPosition(12, 90);
   scroll1->setSizeX(230);
   scroll1->link(interfaceWin);
   scroll1->setMinValue(1.0f);
   scroll1->setMaxValue(360.0f);
   scroll1->setValue(360.0f);
   
   // Steps:
   MVHSCROLLBAR *scroll2 = new MVHSCROLLBAR();   
   scroll2->setPosition(12, 122);
   scroll2->setSizeX(230);
   scroll2->link(interfaceWin);
   scroll2->setMinValue(2.0f);
   scroll2->setMaxValue(200.0f);
   scroll2->setValue(12.0f);    
   
   
      
///////////////
// Mainloop: //
///////////////

   int mouseX = 0, mouseY = 0, oldMouseX = 0, oldMouseY = 0;
   bool mouseBL = false, mouseBR = false;
   bool oldMouseBL = false;
   bool done = false, eventOccurred = false;   
   float rotation = 0.0f, rotationX = 0.0f, rotationY = 0.0f;
   float depth = 300.0f;
      
   ////////////////
   while (!done) //
	{			      	     
	      // Update values:
	      invalid = false;
	      camera->setPosition(0.0f, 0.0f, depth);
	      if (rotate->isClicked())
	            rotation += 15.0f * MVISIO::getFPSfactor();
	      char text[256];
	      interfaceWin->clear();
	      sprintf(text, "^y0005^x0005Points: %d (%d available) \n\n\n^x0004------------------------------\n^x0005Angle: %.1f \n\n^x0005Steps: %d", usedPoints, MAX_POINTS-usedPoints, scroll1->getValue(), (int) scroll2->getValue());
	      interfaceWin->write(text);
	      interfaceWin->format();
	 
	      // Rendering background:	   
	      MVOPENGL::resize(0, 0, MODULE_SIZEX, MODULE_SIZEY);
	   	MVISIO::clear(true, true, false);	   		   	
			MVISIO::begin2D();			   
			   PASS_BACKGROUND;			   			   
			MVISIO::end2D();		
			
			///////////////////
			// LEFT RENDERING:			
			MVOPENGL::resize(0, 0, MODULE_SIZEX/2, MODULE_SIZEY);
			glLineWidth(LINE_WIDTH);
			glPointSize(LINE_WIDTH*2.0);
			MVISIO::begin2D();			      
			      glColor4f(1.0f, 1.0f, 0.0f, 1.0f);
			      glBegin(GL_LINE_STRIP);
			         for (int c=0; c<usedPoints; c++)
			               glVertex2f(point[c].x, point[c].y);
			      glEnd();			      
			      glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
			      glBegin(GL_POINTS);
			         for (int c=0; c<usedPoints; c++)
			               glVertex2f(point[c].x, point[c].y);
			      glEnd();			      
			MVISIO::end2D();			
			
			///////////////////
			// RIGHT RENDERING:
			MVOPENGL::resize(MODULE_SIZEX/2, 0, MODULE_SIZEX/2, MODULE_SIZEY);
			MVISIO::begin3D(camera);				   		   				   
	         glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, (float *) &MVVECTOR4(0.1f, 0.1f, 0.0f, 1.0f));
	         glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, (float *) &MVVECTOR4(0.8f, 0.8f, 0.0f, 1.0f));
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 10.0f);
	         glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, (float *) &MVVECTOR4(0.3f, 0.3f, 0.0f, 1.0f));
	         glEnable(GL_LIGHTING);
	         glEnable(GL_LIGHT0);
	         float tempPosition[] = {0.0f, 20.0f, depth, 1.0f};
	         float tempDirection[] = {0.0f, 0.0f, 0.0f, 1.0f};
	         glLightf(GL_LIGHT0,  GL_SPOT_CUTOFF, 180.0f); 
	         glLightfv(GL_LIGHT0, GL_POSITION, tempPosition);      
            glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, tempDirection);   
            float a[] = {0.1f, 0.1f, 0.1f, 1.0f};
            glLightfv(GL_LIGHT0, GL_AMBIENT, a);
            float d[] = {0.8f, 0.8f, 0.8f, 1.0f};
            glLightfv(GL_LIGHT0, GL_DIFFUSE, d);
            float s[] = {0.2f, 0.2f, 0.2f, 1.0f};
            glLightfv(GL_LIGHT0, GL_SPECULAR, s);
	         glDisable(GL_CULL_FACE);							   
			   glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0f);
			   glRotatef(rotationX, 1.0f, 0.0f, 0.0f);	 						   
			   glRotatef(rotation+rotationY, 0.0f, 1.0f, 0.0f);			   
			   
			   glDisable(GL_LIGHTING);
			   glEnable(GL_BLEND);
			   glBegin(GL_LINE_STRIP);			      
			      glColor4f(1.0f, 1.0f, 1.0f, 0.0f);
			      glVertex3f(0.0f, -60.0f, 0.0f);
			      glColor4f(1.0f, 1.0f, 1.0f, 0.5f);
			      glVertex3f(0.0f, -40.0f, 0.0f);
			      glVertex3f(0.0f, 40.0f, 0.0f);
			      glColor4f(1.0f, 1.0f, 1.0f, 0.0f);
			      glVertex3f(0.0f, 60.0f, 0.0f);			   
			   glEnd();	 			   
			   glDisable(GL_BLEND);
			   glEnable(GL_LIGHTING);			 
			  
            if (wireframe->isClicked())
                  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);					         
			   if (gpoint)
			   {
			         
			         glBegin(GL_TRIANGLES);
			         glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
			         for (int s=1; s<oldSteps; s++)
			               for (int c=0; c<oldUsedPoints-1; c++)
			               {			                     
			                     MVVECTOR3 vA = gpoint[s*oldUsedPoints+c]-gpoint[s*oldUsedPoints+c+1];
			                     MVVECTOR3 vB = gpoint[s*oldUsedPoints+c]-gpoint[(s-1)*oldUsedPoints+c];
			                     MVVECTOR3 n = vA^vB;
			                     n = n.normalize();
			                     glNormal3f(n.x, n.y, n.z);			                     
			                     glVertex3fv((float *) &gpoint[(s-1)*oldUsedPoints+c]);			                           			                     
			                     glVertex3fv((float *) &gpoint[s*oldUsedPoints+c]);			                           			                     			                     
			                     glVertex3fv((float *) &gpoint[s*oldUsedPoints+c+1]);			                           			                     
			                     
			                     glNormal3f(n.x, n.y, n.z);
			                     glVertex3fv((float *) &gpoint[(s-1)*oldUsedPoints+c+1]);			                           			                     
			                     glVertex3fv((float *) &gpoint[(s-1)*oldUsedPoints+c]);		
			                     glVertex3fv((float *) &gpoint[s*oldUsedPoints+c+1]);			                           			                     			                     	                           			                     			                     			                     
			               }
			         glEnd();	
			         
			         // Close volume:
			         if (oldScrollValue<360.0f && oldUsedPoints>2 && oldSteps>1)
			         {			              
			               // First side:
			               MVVECTOR3 vA = gpoint[0]-gpoint[1];
			               MVVECTOR3 vB = gpoint[0]-gpoint[2];
			               MVVECTOR3 n = vA^vB;
			               n = n.normalize();
			               glNormal3f(n.x, n.y, n.z);			
			         
			               GLUtesselator *tess = gluNewTess();
			               gluTessCallback(tess, GLU_TESS_VERTEX, (void (__stdcall*) ()) glVertex3dv);
			               gluTessCallback(tess, GLU_TESS_BEGIN, (void (__stdcall*) ()) beginCallback);
			               gluTessCallback(tess, GLU_TESS_END, (void (__stdcall*) ()) endCallback);
			               gluTessCallback(tess, GLU_TESS_ERROR, (void (__stdcall*) ()) errorCallback);			               
			                          
                        double* v = new double[MAX_POINTS*3];                         
                        for (int c=0; c<oldUsedPoints; c++)
                        {  
                              v[c*3] = gpoint[c].x; 
                              v[c*3+1] = gpoint[c].y; 
                              v[c*3+2] = gpoint[c].z; 
                        }                        
                        gluTessBeginPolygon(tess, NULL); 
                        gluTessBeginContour(tess); 
                        for (int c=0; c<oldUsedPoints; c++)
			               {
			                     gluTessVertex(tess, &v[c*3], &v[c*3]);
			               }
			               gluTessEndContour(tess); 
                        gluTessEndPolygon(tess); 
                        gluDeleteTess(tess);
			               delete [] v;	

                        // Second side:
			               vA = gpoint[(oldSteps-1)*oldUsedPoints]-gpoint[(oldSteps-1)*oldUsedPoints+1];
			               vB = gpoint[(oldSteps-1)*oldUsedPoints]-gpoint[(oldSteps-1)*oldUsedPoints+2];
			               n = vA^vB;
			               n = n.normalize();
			               glNormal3f(n.x, n.y, n.z);			
			         
			               tess = gluNewTess();
			               gluTessCallback(tess, GLU_TESS_VERTEX, (void (__stdcall*) ()) glVertex3dv);
			               gluTessCallback(tess, GLU_TESS_BEGIN, (void (__stdcall*) ()) beginCallback);
			               gluTessCallback(tess, GLU_TESS_END, (void (__stdcall*) ()) endCallback);
			               gluTessCallback(tess, GLU_TESS_ERROR, (void (__stdcall*) ()) errorCallback);
			                          
                        v = new double[MAX_POINTS*3];                                                 
                        for (int c=0; c<oldUsedPoints; c++)
                        {  
                              v[c*3] = gpoint[(oldSteps-1)*oldUsedPoints+c].x; 
                              v[c*3+1] = gpoint[(oldSteps-1)*oldUsedPoints+c].y; 
                              v[c*3+2] = gpoint[(oldSteps-1)*oldUsedPoints+c].z; 
                        }                        
                        gluTessBeginPolygon(tess, NULL); 
                        gluTessBeginContour(tess); 
                        for (int c=0; c<oldUsedPoints; c++)
			               {
			                     gluTessVertex(tess, &v[c*3], &v[c*3]);
			               }
			               gluTessEndContour(tess); 
                        gluTessEndPolygon(tess); 
                        gluDeleteTess(tess);
			               delete [] v;				             
			         }		         		         
			   }
			   if (wireframe->isClicked())
                  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);					         
			   glDisable(GL_LIGHTING);
			MVISIO::end3D();					
			
			//////////////
			// Foreground:						
			MVOPENGL::resize(0, 0, MODULE_SIZEX, MODULE_SIZEY);
			MVISIO::begin2D();	
			   // Separator:
			   glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			   glBegin(GL_LINE_STRIP);
			         glVertex2f(left_ex, left_sy);
			         glVertex2f(left_ex, left_ey);
			         glVertex2f(right_sx, right_ey);
			         glVertex2f(right_sx, right_sy);
			         glVertex2f(left_ex, left_sy);
			   glEnd();			 	
			   
			   // Title:
			   font->putString(10, 13, MODULE_NAME);	 
            if (invalid && oldScrollValue < 360.0f)			   
			         font->putString((int) right_sx+ ( (right_ex-right_sx)/2.0)-font->stringLength("Invalid shape")/2.0f, MODULE_SIZEY/2, "^c8228Invalid shape"); 
			   
			   // Interface:
			   PASS_QUITBUTTON;			 		  			  			   
			   plane->pass();		  				 			   
			   interfaceWin->pass();
			MVISIO::end2D();
			
			
			////////////////
			// Check events:			
			SDL_Event _event;			
			//SDL_PumpEvents();						
			while (SDL_PollEvent(&_event)) 
			{
					// Quit by click on X:
					if (_event.type == SDL_QUIT) 
							done = true;	
               
               // Quit by pressing ESC:
					if (_event.type == SDL_KEYDOWN) 
					{
							// Quit by ESC:
							if (_event.key.keysym.sym == SDLK_ESCAPE) 
							       done = true;															
			      }															
			}        
			
			// Update keyboard (like a joypad):
			Uint8 *_array;
			int arraySize;
			_array = SDL_GetKeyState(&arraySize);
			if (_array[SDLK_PAGEUP])    
			{
               depth += 100.0f * MVISIO::getFPSfactor();  
               if (depth > 400.0f)
			            depth = 400.0f;			             
         }							       								                     
         else
               if (_array[SDLK_PAGEDOWN])
               {
                     depth -= 100.0f * MVISIO::getFPSfactor();
                       if (depth < 150.0f)
			                  depth = 150.0f;                                          
               }							     			
			
			float deltaX = (float) (mouseX - oldMouseX);
         float deltaY = (float) (mouseY - oldMouseY);		
         oldMouseX = mouseX; oldMouseY = mouseY;	
         oldMouseBL = mouseBL;	
			
		   // Update mouse:
			int buttonState = SDL_GetMouseState(&mouseX, &mouseY);		
         if (buttonState&SDL_BUTTON(1))
					mouseBL = true;					
			else 
					mouseBL = false;
         if (buttonState&SDL_BUTTON(2) || buttonState&SDL_BUTTON(3))
					mouseBR = true;					
			else 
					mouseBR = false;					
			
			// Check MVisio events:
			MVID id=0;
			eventOccurred = MVELEMENT::manageGUIEvents(0, mouseX, mouseY, mouseBL, mouseBR, &id);  
			if (PRESSED_QUITBUTTON(id))
			      done = true;
        
         if (id==reset->getId())
               usedPoints = 0;			      
               
         if (id==apply->getId())
         {
               if (createSurface(point, &gpoint, usedPoints, (int) scroll2->getValue()+1.0f, scroll1->getValue(), (float) plane->getPosition().x))
               {
                     oldUsedPoints = usedPoints;  
                     oldSteps = (int) scroll2->getValue()+1;    
                     oldScrollValue = scroll1->getValue();               
               }
         }               
			      
         // Create a new point:
         if (!eventOccurred)
               if (mouseBL==true && !oldMouseBL && usedPoints < MAX_POINTS)
                     if (mouseX>=left_sx && mouseY>=left_sy && mouseX<=left_ex && mouseY<=left_ey)
                     {                                       
                           point[usedPoints].x = mouseX;
                           point[usedPoints].y = mouseY;
                           for (int c=0; c<usedPoints; c++)
                                 if (point[c].distance(MVVECTOR2(mouseX, mouseY)) <= 4.0f)
                                 {
                                       point[usedPoints] = point[c];
                                 }                         
                           usedPoints++;
                     }         			      
         
         // Rotate scene:
         if (!eventOccurred && mouseX>MODULE_SIZEX/2)
         {
               if (mouseBL && !mouseBR)
               {
                     // Rotate model:
			            rotationX += deltaY/3.5f;
			            rotationY += deltaX/3.5f;
			            if (rotationX > 80.0f) rotationX = 80.0f;
			            if (rotationX < -80.0f) rotationX = -80.0f;			         
			      }  
			      if (mouseBL && mouseBR)
			      {
			            depth += deltaY;
			            if (depth < 150.0f)
			                  depth = 150.0f;
                     if (depth > 400.0f)
			                  depth = 400.0f;			                  			               
			      }                   
         }			   
         			
			// Display:		
			MVISIO::swap();		
			Sleep(10); // Don't kill mthreading ;)
	}      
   
   // Done:
   if (gpoint)
         delete [] gpoint;
   MVISIO::free();
   return 0;
}
